"""
    saraki.auth
    ~~~~~~~~~~~
"""

import jwt
from functools import wraps
from datetime import datetime

from cerberus import Validator
from flask import request, current_app, jsonify, abort, _request_ctx_stack
from werkzeug.routing import BaseConverter
from werkzeug.local import LocalProxy

from .utility import generate_schema, get_key_path
from .model import (
    User,
    Org,
    Membership,
    _persist_actions,
    _persist_resources,
    _persist_abilities,
    get_member_privileges,
)
from .exc import (
    NotFoundCredentialError,
    InvalidUserError,
    InvalidOrgError,
    InvalidMemberError,
    InvalidPasswordError,
    JWTError,
    TokenNotFoundError,
    AuthorizationError,
    ProgrammingError,
)


AUTH_SCHEMA = generate_schema(User, include=["username", "password"])


HTTP_VERBS_CRUD = {
    "get": "read",
    "post": "write",
    "patch": "write",
    "put": "write",
    "delete": "delete",
}


#: A local proxy object that points to the user accessing an endpoint in the
#: current request. The value of this object is an instance of the model class
#: :class:`~saraki.model.User` or None if there is not a user.
current_user = LocalProxy(lambda: getattr(_request_ctx_stack.top, "current_user", None))


#: A local proxy object that points to the organization being accessed in the
#: current request. The value of this object is an instance of the model class
#: :class:`~saraki.model.Org` or None if the endpoint is not a tenant endpoint.
current_org = LocalProxy(lambda: getattr(_request_ctx_stack.top, "current_org", None))


class Claim(str):
    def __new__(cls, value, type):
        return str.__new__(cls, value)

    def __init__(self, value, type):
        self._claim_type = type

    @property
    def type(self):
        return self._claim_type

    def __repr__(self):
        return f'Claim(value="{self[:]}", type="{self.type}")'


class SubClaimConverter(BaseConverter):
    def to_python(self, value):
        return Claim(value=value, type="sub")

    def to_url(self, value):
        return value


class AudClaimConverter(BaseConverter):
    def to_python(self, value):
        return Claim(value=value, type="aud")

    def to_url(self, value):
        return value


def _verify_username(username):

    filters = {"canonical_username": username.lower()}
    identity = User.query.filter_by(**filters).one_or_none()

    if identity is None:
        raise InvalidUserError(f'Username "{username}" is not registered')

    return identity


def _verify_orgname(orgname):

    org = Org.query.filter_by(orgname=orgname).one_or_none()

    if org is None:
        raise InvalidOrgError(f'Orgname "{orgname}" is not registered')

    return org


def _verify_member(user, org):
    member = Membership.query.filter_by(user_id=user.id, org_id=org.id).one_or_none()

    if member is None:
        raise InvalidMemberError(f"{user.username} is not a member of {org.orgname}")

    return member


def _get_request_jwt():
    """Return ``Authorization`` header token if present, otherwise None.

    If the Authorization header is present, raises a JWTError if the token
    is malformed.
    """

    token_string = request.headers.get("Authorization", None)
    token_prefix = current_app.config["JWT_AUTH_HEADER_PREFIX"]

    if token_string is None:
        return None

    parts = token_string.split()

    error = (
        "Missing or malformed token"
        if len(parts) < 2
        else "The token contains spaces"
        if len(parts) > 2
        else "Unsupported authorization type"
        if parts[0] != token_prefix
        else None
    )

    if error:
        raise JWTError(error)

    return parts[1]


def _generate_jwt_payload(user, org=None):
    required_claim_list = current_app.config["JWT_REQUIRED_CLAIMS"]
    iat = datetime.utcnow()
    exp = iat + current_app.config["JWT_EXPIRATION_DELTA"]
    iss = current_app.config["JWT_ISSUER"] or current_app.config["SERVER_NAME"]
    payload = {}

    if "iss" in required_claim_list:
        if not iss:
            raise RuntimeError(
                "The token payload could not be generated. The claim iss is "
                "required, but neither JWT_ISSUER nor SERVER_NAME are provided"
            )

        payload["iss"] = iss

    if org:
        payload["aud"] = org.orgname
        payload.update({"scp": get_member_privileges(org, user)})

    payload.update({"iat": iat, "exp": exp, "sub": user.username})

    return payload


def _encode_jwt(payload):
    secret = current_app.config["SECRET_KEY"]

    if secret is None:
        raise RuntimeError("SECRET_KEY is not set. Can not generate Token")

    algorithm = current_app.config["JWT_ALGORITHM"]
    required_claim_list = current_app.config["JWT_REQUIRED_CLAIMS"]

    missing_claims = list(set(required_claim_list) - set(payload.keys()))

    if missing_claims:
        raise ValueError(
            f'Payload is missing required claims: {", ".join(missing_claims)}'
        )

    return jwt.encode(payload, secret, algorithm=algorithm)


def _decode_jwt(token):

    if not isinstance(token, (str, bytes)):
        raise ValueError(f"{type(token)} is not a valid JWT string")

    required_claim_list = current_app.config["JWT_REQUIRED_CLAIMS"]

    options = {"require_" + claim: True for claim in required_claim_list}
    options.update({"verify_" + claim: True for claim in required_claim_list})
    options["verify_aud"] = False

    parameters = {
        "jwt": token,
        "key": current_app.config["SECRET_KEY"],
        "leeway": current_app.config["JWT_LEEWAY"],
        "options": options,
        "algorithms": [current_app.config["JWT_ALGORITHM"]],
    }

    if "iss" in required_claim_list:
        parameters["issuer"] = current_app.config["JWT_ISSUER"]

    try:
        payload = jwt.decode(**parameters)
    except jwt.exceptions.MissingRequiredClaimError as e:
        raise JWTError(str(e))
    except jwt.exceptions.ExpiredSignatureError as e:
        raise JWTError("Token has expired")
    except jwt.exceptions.InvalidIssuerError as e:
        raise JWTError(str(e))
    except jwt.exceptions.DecodeError as e:
        raise JWTError("Invalid or malformed token")

    return payload


def _get_parent_resource(required_resource, scopes):

    resource_map = current_app.auth.resources
    path = get_key_path(required_resource, resource_map) or []

    for resource in scopes:
        if resource in path:
            return resource

    return None


def _is_authorized(payload, resource=None, action=None):

    criteria = []

    for c in request.view_args.values():
        if isinstance(c, Claim):
            criteria.append(c.type in payload and c == payload[c.type])

    if not all(criteria):
        return False

    if not resource:
        return True

    scopes = payload.get("scp")

    if not scopes:
        return False

    if action is None:
        method = request.method.lower()

        if method not in HTTP_VERBS_CRUD:
            return False

        action = HTTP_VERBS_CRUD[method]

    scope_resource = scopes.get(resource)

    if not scope_resource:
        parent_resource = _get_parent_resource(resource, scopes)

        if not parent_resource:
            return False

        scope_resource = scopes[parent_resource]

    return action in scope_resource or "manage" in scope_resource


def _validate_request(resource=None, action=None):

    token = _get_request_jwt()

    if token is None:
        raise TokenNotFoundError

    payload = _decode_jwt(token)

    if _is_authorized(payload, resource, action) is False:
        raise AuthorizationError

    org = None

    try:
        user = _verify_username(payload["sub"])
        org = _verify_orgname(payload["aud"]) if "aud" in payload else None

    except (InvalidUserError, InvalidOrgError) as e:
        raise AuthorizationError

    _request_ctx_stack.top.current_user = user
    _request_ctx_stack.top.current_org = org


# ~~~~~~~~~~~~~~~~~~~~~
#
# AUTHENTICATION
#
# ~~~~~~~~~~~~~~~~~~~~~


def _authenticate_with_token(token):
    """Given a valid access token, authenticate a user and return a new access
    token.
    """

    payload = _decode_jwt(token)
    username = payload["sub"]

    return _verify_username(username)


def _authenticate_with_password(username, password):

    user = _verify_username(username)

    if user.verify_password(password) is False:
        raise InvalidPasswordError("Invalid password")

    return user


def _authenticate(orgname=None):
    """Handles an authentication request and returns an access token."""

    org = None

    if orgname:
        try:
            org = _verify_orgname(orgname)
        except InvalidOrgError:
            raise abort(404)

    token = _get_request_jwt()

    if token:
        user = _authenticate_with_token(token)
    else:
        username_password = request.get_json()

        if username_password is None:
            raise NotFoundCredentialError("Missing token and username/password")

        v = Validator(AUTH_SCHEMA)

        if v.validate(username_password) is False:
            abort(400, v.errors)

        user = _authenticate_with_password(**username_password)

    if org:
        _verify_member(user, org)

    payload = _generate_jwt_payload(user, org)
    access_token = _encode_jwt(payload)

    return jsonify({"access_token": access_token.decode("utf-8")})


def require_auth(resource=None, action=None, parent_resource=None):

    if action and not resource:
        raise ProgrammingError(f"You passed an action '{action}' without a resource")

    def decorator(func):

        func._auth_metadata = dict(
            resource=resource, action=action, parent_resource=parent_resource
        )

        @wraps(func)
        def wrapper(*arg, **karg):

            _validate_request(resource, action)

            return func(*arg, **karg)

        return wrapper

    return decorator


class Auth:
    def __init__(self, app=None):

        self._resources = {}
        self._actions = ["manage"]
        self._persist_actions_func = _persist_actions
        self._persist_resources_func = _persist_resources
        self._persist_abilities_func = _persist_abilities

        if app:
            self.init_app(app)

    def init_app(self, app):

        self.app = app

        app.url_map.converters["sub"] = SubClaimConverter
        app.url_map.converters["aud"] = AudClaimConverter

        methods = ["POST"]

        app.add_url_rule(
            rule="/auth",
            view_func=_authenticate,
            methods=methods,
            defaults={"orgname": None},
        )

        app.add_url_rule(
            rule="/auth/<orgname>", view_func=_authenticate, methods=methods
        )

    @property
    def resources(self):
        return self._resources

    @property
    def actions(self):
        return self._actions

    def _add_resource(self, resource, parent=None):

        if type(resource) != str:
            raise TypeError(
                f"resource argument must be an string, got {type(resource)}"
            )

        resources = self.resources
        path = get_key_path(resource, resources)

        if not path:

            parent_path = get_key_path(parent, resources) if parent else []

            if parent_path is None:
                self._resources[parent] = None
                parent_path = [parent]

            parent_resource = self._resources

            for branch in parent_path:
                if parent_resource[branch] is None:
                    parent_resource[branch] = {}
                parent_resource = parent_resource[branch]

            parent_resource[resource] = None

    def _add_action(self, action):

        if type(action) != str:
            raise TypeError(f"action argument must be an string, got {type(action)}")
        if action is not None and action not in self.actions:
            self._actions.append(action)

    def _collect_metadata(self):

        for view_func in self.app.view_functions.values():

            if hasattr(view_func, "_auth_metadata"):

                _auth = view_func._auth_metadata

                resource = _auth["resource"]
                parent_resource = _auth["parent_resource"]
                action = _auth["action"]

                if resource:
                    self._add_resource(resource, parent_resource)

                if action:
                    self._add_action(action)

    def persist_actions(self, f):
        """Registers a function called to persist actions in a database."""

        self._persist_actions_func = f
        return f

    def persist_resources(self, f):
        """Registers a function called to persist resources in a database."""

        self._persist_resources_func = f
        return f

    def persist_abilities(self, f):
        """Registers a function called to persist abilities in a database."""

        self._persist_resources_func = f
        return f

    def persist_data(self):
        self._collect_metadata()
        self._persist_actions_func(self.actions)
        self._persist_resources_func(self.resources)
        self._persist_abilities_func()



import pytest
from unittest.mock import patch
from json import loads
from sqlalchemy.orm import joinedload
from werkzeug.exceptions import UnsupportedMediaType, BadRequest
from flask import make_response
from saraki.utility import (
    export_from_sqla_object,
    ExportData,
    import_into_sqla_object,
    json,
    Validator,
    get_key_path,
)
from common import Cartoon, Product, Order, OrderLine, DummyBaseModel, DummyModel, Todo


def test_import_into_sqla_object():
    product = Product()
    data = {"name": "Acme anvils", "color": "black", "price": 99}

    import_into_sqla_object(product, data)

    assert product.id is None
    assert product.name == "Acme anvils"
    assert product.color == "black"
    assert product.price == 99


@pytest.mark.usefixtures("data")
class TestExportData:
    def test_with_invalid_arguments(self):
        message = "Pass a valid SQLAlchemy mapped class instance"
        export_data = ExportData()

        class NonMappedModel(object):
            pass

        with pytest.raises(ValueError, match=message):
            export_data(NonMappedModel())

        with pytest.raises(ValueError, match=message):
            export_data(1)

    def test_non_persisted_untouched_object(self):
        export_data = ExportData()

        data = export_data(Product())
        assert len(data) == 7

        for prop in ["id", "name", "created_at", "updated_at"]:
            assert data[prop] is None

        assert data["color"] == "white"
        assert data["price"] == 0
        assert data["enabled"] is False

    def test_non_persisted_touched_object(self):
        export_data = ExportData()

        data = export_data(Product(id=9, name="TNT", price=799, color="red"))

        assert len(data) == 7
        assert data["id"] == 9
        assert data["name"] == "TNT"
        assert data["color"] == "red"
        assert data["price"] == 799
        assert data["enabled"] is False
        assert data["created_at"] is None
        assert data["updated_at"] is None

    def test_persisted_object(self, ctx):
        export_data = ExportData()
        product = Product.query.get(1)
        data = export_data(product)

        assert len(data) == 7
        assert data["id"] == 1
        assert data["name"] == "Explosive Tennis Balls"
        assert data["color"] == "white"
        assert data["price"] == 9

    def test_loaded_one_to_many_relationship(self, ctx):
        export_data = ExportData()

        order = Order.query.options(joinedload(Order.lines)).get(1)
        data = export_data(order)

        assert len(data) == 3
        assert len(data["lines"]) == 3
        assert {"product_id": 1, "quantity": 3, "unit_price": 14} in data["lines"]
        assert {"product_id": 2, "quantity": 4, "unit_price": 142} in data["lines"]
        assert {"product_id": 3, "quantity": 7, "unit_price": 73} in data["lines"]

    def test_loaded_one_to_one_relationship(self, ctx):
        export_data = ExportData()

        order_line = OrderLine.query.options(joinedload(OrderLine.product)).get((1, 2))
        data = export_data(order_line)

        assert len(data) == 5
        assert data["order_id"] == 1
        assert data["product_id"] == 2

        product_data = data["product"]
        assert len(product_data) == 7
        assert product_data["id"] == 2
        assert product_data["name"] == "Binocular"
        assert product_data["color"] == "black"
        assert product_data["price"] == 99

    def test_explicit_column_inclusion(self, ctx):
        export_data = ExportData()

        product = Product.query.get(1)
        data = export_data(product, include=["id", "name"])

        assert len(data) == 2
        assert data["id"] == 1
        assert data["name"] == "Explosive Tennis Balls"

    def test_explicit_column_exclusion(self, ctx):
        export_data = ExportData()

        product = Product.query.get(1)
        data = export_data(product, exclude=["id", "name"])

        assert len(data) == 5
        assert "id" not in data
        assert "name" not in data

    def test_explicit_loaded_relationship_exclusion(self, ctx):
        export_data = ExportData()

        order = (
            Order.query.options(joinedload(Order.lines))
            .options(joinedload(Order.customer))
            .get(1)
        )

        data = export_data(order, exclude=["lines"])

        assert "lines" not in data
        assert "customer" in data

    def test_include_and_exclude_argument(self, ctx):
        export_data = ExportData()
        product = Product.query.get(1)

        data = export_data(product, include=["id", "name", "price"], exclude=["id"])

        assert len(data) == 2
        assert "id" not in data

    def test_exclude_property(self, ctx):
        export_data = ExportData(exclude=["id"])
        product = Product.query.get(1)

        data = export_data(product)

        assert "id" not in data

    def test_exclude_property_on_model_with_export_data_method(self, ctx):
        assert hasattr(Cartoon, "export_data")
        assert hasattr(Cartoon, "id")

        export_data = ExportData(exclude=["id"])
        product = Cartoon.query.first()
        data = export_data(product)

        assert "id" not in data

    def test_exclude_property_on_list_of_models_with_export_data_method(self, ctx):
        assert hasattr(Todo, "export_data")

        lst = Cartoon.query.all()
        assert len(lst) > 0

        output = export_from_sqla_object(lst)
        assert all(["org_id" not in model for model in output])

    def test_exclude_property_on_list_of_models(self, ctx):
        export_data = ExportData(exclude=["customer_id"])

        assert not hasattr(Order, "export_data")
        lst = Order.query.all()

        assert len(lst) > 0
        output = export_data(lst)

        assert all(["customer_id" not in model for model in output])

    def test_exclude_property_and_include_parameter(self, ctx):
        export_data = ExportData(exclude=["id"])
        product = Product.query.get(1)

        data = export_data(product, include=["id"])

        assert "id" not in data

    def test_list_of_models_without_export_data(self, ctx):
        export_data = ExportData()
        lst = export_data(Order.query.all())

        assert type(lst) == list
        assert len(lst) == 2
        assert {"id": 1, "customer_id": 1} in lst

    def test_list_of_models_with_export_method(self, ctx):
        export_data = ExportData()
        lst = export_data(OrderLine.query.all())
        data = lst[0]

        assert len(data) == 3
        assert "product_id" in data
        assert "unit_price" in data
        assert "quantity" in data

    def test_explicit_inclusion_on_list_of_models(self, ctx):
        export_data = ExportData()
        lst = export_data(Order.query.all(), include=["id"])

        assert type(lst) == list
        assert len(lst) == 2
        assert all(["customer_id" not in model for model in lst])

    def test_explicit_exclusion_on_list_of_models(self, ctx):
        export_data = ExportData()
        lst = export_data(Order.query.all(), exclude=["customer_id"])

        assert type(lst) == list
        assert len(lst) == 2
        assert all(["customer_id" not in model for model in lst])

    def test_explicit_inclusion_on_list_of_model_with_export_data_method(self, ctx):
        export_data = ExportData()
        lst = Cartoon.query.all()
        output = export_data(Cartoon.query.all(), include=["id"])

        assert len(output) == len(lst)
        assert all(["name" not in model for model in output])

    def test_explicit_exclusion_on_list_of_models_with_export_data_method(self, ctx):
        export_data = ExportData()
        lst = OrderLine.query.all()
        output = export_data(lst, exclude=["product_id"])

        assert len(output) == len(lst)
        assert all(["product_id" not in model for model in output])

    def test_loaded_one_to_one_relationship_with_export_data_method(self, ctx):
        export_data = ExportData()

        order_line = Order.query.options(joinedload(Order.customer)).get(1)
        data = export_data(order_line, include=["id"])

        assert len(data) == 2

        customer = data["customer"]
        assert len(customer) == 2
        assert "id" in customer
        assert "firstname" in customer

    def test_model_class_export_data_with_exception(self, request_ctx):
        class Fake:
            def export_data(self, include, exclude):
                raise AttributeError("Inside of export_data method")

        with pytest.raises(AttributeError, match="Inside of export_data method"):
            export_from_sqla_object([Fake()])

    def test_passed_arguments_on_list_of_model_with_export_data_method(self, ctx):
        """Test that the global list of excluded columns are not passed down
        to export_data method.
        """

        with patch.object(OrderLine, "export_data"):
            model = OrderLine()
            export_from_sqla_object([model], exclude=["product_id"])

            model.export_data.assert_called_with((), ["product_id"])


@pytest.mark.wip
class TestJson:
    @pytest.mark.parametrize(
        "returned, expected",
        [
            ({"id": 1}, {"id": 1}),
            ([1, 2, 3], [1, 2, 3]),
            (DummyBaseModel(id=2), {"id": 2}),
            (DummyModel(id=2), {"id": 2}),
            ("Hello", "Hello"),
            (1, 1),
            (None, None),
        ],
    )
    def test_return_single_objects(self, returned, expected, request_ctx):
        @json
        def view_func():
            return returned

        with request_ctx("/"):
            rv = view_func()

        assert rv.status_code == 200
        assert rv.content_type == "application/json"
        assert loads(rv.data) == expected

    @pytest.mark.parametrize(
        "returned, expected",
        [
            (({"id": 1}, 500), (500, {"id": 1})),
            (([1, 2, 3], 404), (404, [1, 2, 3])),
            (("Hello", 201), (201, "Hello")),
            ((1, 201), (201, 1)),
            ((None, 201), (201, None)),
            ((DummyBaseModel(id=2), 201), (201, {"id": 2})),
            ((DummyModel(id=2), 201), (201, {"id": 2})),
            (({"id": 14}, {"X-Header": "value"}), (200, {"id": 14})),
        ],
        ids=[
            "dict type, status",
            "list type, status",
            "string type, status",
            "integer type, status",
            "None type, status",
            "model without export_data, status",
            "model with export_data, status",
            "dict, headers",
        ],
    )
    def test_status_code(self, returned, expected, request_ctx):
        view_func = json(lambda: returned)

        with request_ctx("/"):
            rv = view_func()

        assert rv.status_code == expected[0]
        assert rv.content_type == "application/json"
        assert loads(rv.data) == expected[1]

    @pytest.mark.parametrize(
        "returned, expected",
        [
            (({"id": 1}, 500, {"X-header": "x-value"}), (500, {"id": 1})),
            (([1, 2, 3], 404, {"X-header": "x-value"}), (404, [1, 2, 3])),
            ((DummyBaseModel(id=2), 201, {"X-header": "x-value"}), (201, {"id": 2})),
            ((DummyModel(id=2), 201, {"X-header": "x-value"}), (201, {"id": 2})),
            (("Hello", 201, {"X-header": "x-value"}), (201, "Hello")),
            ((1, 201, {"X-header": "x-value"}), (201, 1)),
            ((None, 201, {"X-header": "x-value"}), (201, None)),
        ],
    )
    def test_return_extra_http_header(self, returned, expected, request_ctx):
        @json
        def view_func():
            return returned

        with request_ctx():
            rv = view_func()

        assert rv.status_code == expected[0]
        assert loads(rv.data) == expected[1]
        assert rv.content_type == "application/json"
        assert rv.headers["X-header"] == "x-value"

    def test_return_custom_response(self, request_ctx):
        @json
        def index():
            return make_response("Hello world", 201)

        with request_ctx("/"):
            rv = index()

        assert rv.status_code == 201
        assert rv.data == b"Hello world"
        assert rv.content_type != "application/json"

    def test_post_request_with_wrong_content_type(self, request_ctx):
        @json
        def view_func():
            pass

        error_message = "application/json mimetype expected"

        with request_ctx(method="POST", content_type="text/plain"):
            with pytest.raises(UnsupportedMediaType, match=error_message):
                view_func()

    def test_post_request_with_invalid_json_object(self, request_ctx):
        @json
        def view_func():
            pass

        error_message = "The request payload has an invalid JSON object"
        params = {
            "method": "POST",
            "content_type": "application/json",
            "data": "{'prop': 'value'}",
        }

        with request_ctx(**params):
            with pytest.raises(BadRequest, match=error_message):
                view_func()

    def test_post_request_with_valid_json_object(self, request_ctx):
        @json
        def view_func():
            pass

        config = {"method": "POST", "content_type": "application/json", "data": "{}"}

        with request_ctx(**config):
            view_func()

    @patch("saraki.utility.is_sqla_obj", return_value=True)
    def test_model_class_export_data_with_exception(self, is_sqla_obj, request_ctx):
        class Fake:
            def export_data(self, include=None, exclude=None):
                raise AttributeError("Inside of export_data method")

        view_func = json(lambda: Fake())

        with request_ctx("/"):
            with pytest.raises(AttributeError, match="Inside of export_data method"):
                view_func()


@pytest.mark.usefixtures("data")
class TestValidator:
    def test_constructor(self):
        v = Validator({}, Order)
        assert v.model_class is Order

    def test_unique_rule_missing_model_class(self, ctx):
        v = Validator({"name": {"unique": True}})

        with pytest.raises(RuntimeError):
            v.validate({"name": "Update name"})

    def test_unique_rule(self, ctx):
        v = Validator({"name": {"unique": True}}, Product)
        assert v.validate({"name": "New product name"}) is True

    def test_unique_rule_when_name_already_exist(self, ctx):
        v = Validator({"name": {"unique": True}}, Product)
        assert v.validate({"name": "Acme anvils"}) is False
        assert "Must be unique, but 'Acme anvils' already exist" in v.errors["name"]

    def test_unique_rule_update_without_model_argument(self, ctx):
        v = Validator({"name": {"unique": True}}, Product)

        assert v.validate({"name": "Update name"}, update=True) is True
        assert v.validate({"name": "Acme anvils"}, update=True) is False

    def test_unique_rule_update_with_model_argument(self, ctx):
        model = Product.query.filter_by(name="Acme anvils").one()

        v = Validator({"name": {"unique": True}}, Product)

        kargs = {"update": True, "model": model}
        assert v.validate({"name": "Acme anvils"}, **kargs) is True
        assert v.validate({"name": "new product name"}, **kargs) is True
        assert v.validate({"name": "Binocular"}, **kargs) is False

        error_message = "Must be unique, but 'Binocular' already exist"
        assert error_message in v.errors["name"]


@pytest.fixture(scope="module")
def _object():
    return {
        "5": None,
        "1": {"2": None, "3": {"7": None, "4": {"0": None}}},
        "10": {"11": None, "12": {"13": {"14": None}, "15": None}},
        "6": None,
    }


@pytest.mark.parametrize(
    "key, result",
    [
        ("9", None),
        ("5", ["5"]),
        ("6", ["6"]),
        ("3", ["1", "3"]),
        ("4", ["1", "3", "4"]),
        ("14", ["10", "12", "13", "14"]),
        ("15", ["10", "12", "15"]),
    ],
)
def test_get_key_path(key, _object, result):
    assert get_key_path(key, _object) == result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from configparser import ConfigParser
from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()


def clean(k, v):
    rk = k[1:-1] if k[0] == '"' else k
    rv = v[1:-1] if v[0] == '"' else v
    return rk, rv


# Pipfile is in toml format and the only parser available in installation time
# is ConfigParser from the standar library. It can sort of parse it with some
# help.


pipfile = ConfigParser()
pipfile.read("Pipfile")
packages = dict(pipfile["packages"])

requirements = ["".join(clean(key, value)) for key, value in packages.items()]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest"]

setup(
    author="José María Domínguez Moreno",
    author_email="miso.0b11@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
    ],
    description="A web application helper",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="saraki",
    name="saraki",
    packages=find_packages(include=["saraki"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/mbarakaja/saraki",
    version="0.1.0a0",
    zip_safe=False,
)



def is_list_subset_another(subset_lst, lst):

    if subset_lst is lst:
        return True

    criteria = []

    for subitem in subset_lst:
        subset = subitem.items()
        item_is_subset = False

        for item in lst:
            if subset <= item.items():
                item_is_subset = True
                break

        criteria.append(item_is_subset)

    return all(criteria)


class AssertList:
    def __init__(self, lst):
        self.lst = lst

    def subset_of(self, other_lst):
        return is_list_subset_another(self.lst, other_lst)

    def has_subset(self, other_lst):
        return is_list_subset_another(other_lst, self.lst)

    def __ge__(self, other_list):
        return self.has_subset(other_list)

    def __le__(self, other_list):
        return self.subset_of(other_list)


def pytest_assertrepr_compare(config, op, left, right):

    if isinstance(left, AssertList) and op in {"<=", ">="}:

        statement = f"{left.lst} {op} {right}"

        return [statement]


list_is = AssertList  # noqa: F841



from flask import jsonify, request, abort
from sqlalchemy.orm import joinedload
from saraki import Blueprint
from .auth import require_auth, current_user, current_org
from .utility import generate_schema, json, export_from_sqla_object, Validator
from .model import (
    database,
    Plan,
    User,
    Org,
    Membership,
    Resource,
    Action,
    Ability,
    Role,
    RoleAbility,
    MemberRole,
)


user_schema = generate_schema(User, exclude=["canonical_username", "active"])


def signup_view():

    data = request.get_json()

    v = Validator(user_schema)

    if v.validate(data) is False:
        abort(400, v.errors)

    user = User()
    user.import_data(data)

    database.session.add(user)
    database.session.commit()

    return jsonify({"username": user.username}), 201


appbp = Blueprint("app", __name__)


# Resources
# ~~~~~~~~~

appbp.add_resource(
    Resource,
    "resources",
    methods={"item": ["GET"], "list": ["GET"]},
    resource_name="app",
)


# Action
# ~~~~~~

appbp.add_resource(
    Action, "actions", methods={"item": ["GET"], "list": ["GET"]}, resource_name="app"
)


# Ability
# ~~~~~~

appbp.add_resource(
    Ability,
    "abilities",
    methods={"item": ["GET"], "list": ["GET"]},
    resource_name="app",
)


#
# Application plans
# ~~~~~~~~~~~~~~~~~
#

appbp.add_resource(
    Plan, "plans", methods={"item": ["GET"], "list": ["GET"]}, secure=False
)

appbp.add_resource(
    Plan,
    "plans",
    resource_name="app",
    methods={"item": ["PATCH", "DELETE"], "list": ["POST"]},
)

"""
    User organizations
    ~~~~~~~~~~~~~~~~~~
"""

ORG_SCHEMA = generate_schema(Org, exclude=["id", "user_id"])
ORG_SCHEMA["orgname"]["unique"] = True


def _add_member(org, user, extra_data=None):
    data = {"user_id": user.id, "org_id": org.id}

    if extra_data:
        data.update(extra_data)

    member = Membership()
    member.import_data(data)
    database.session.add(member)

    return member


@appbp.route("/users/<sub:username>/orgs")
@require_auth()
@json
def list_user_organizations(username):
    """Return a list of organization accounts of a user. This includes those
    owned by the user and those where the user is a member.
    """

    user_id = current_user.id

    memberships = Membership.query.filter_by(user_id=user_id).all()

    org_list = [export_from_sqla_object(m.org) for m in memberships]

    return org_list, 200


@appbp.route("/users/<sub:username>/orgs", methods=["POST"])
@require_auth()
@json
def add_organization_account(username):
    """Creates an new organization account.

    When an user creates an organization account, this user is automatically
    added to the list of members of the organization and then flagged as the
    owner.
    """

    data = request.get_json()
    v = Validator(ORG_SCHEMA, Org)

    if v.validate(data) is False:
        abort(400, v.errors)

    user = current_user._get_current_object()

    data["user_id"] = current_user.id

    org = Org()
    org.import_data(data)

    database.session.add(org)
    database.session.flush()

    _add_member(org, user, {"is_owner": True})

    database.session.commit()

    return org, 201


@appbp.route("/orgs/<aud:orgname>/members")
@require_auth("org")
@json
def list_members(orgname):
    org_id = current_org.id
    member_list = (
        Membership.query.options(joinedload("user")).filter_by(org_id=org_id).all()
    )

    return [member.export_data() for member in member_list], 200


def member_username_validator(field, value, error):
    username = value
    user = User.query.filter_by(username=value).one_or_none()

    if not user:
        error(field, f"User {username} does not exist")
        return

    member = Membership.query.filter_by(
        user_id=user.id, org_id=current_org.id
    ).one_or_none()

    if member:
        orgname = current_org.orgname
        error(field, f"{username} is already a member of {orgname}")


new_member_schema = {
    "username": {
        "type": "string",
        "validator": member_username_validator,
        "required": True,
    }
}


@appbp.route("/orgs/<aud:orgname>/members", methods=["POST"])
@require_auth("org")
@json
def add_member(orgname):
    data = request.get_json()
    v = Validator(new_member_schema)

    if v.validate(data) is False:
        abort(400, v.errors)

    username = data["username"]
    user = User.query.filter_by(username=username).one()

    member = Membership(user_id=user.id, org_id=current_org.id)

    database.session.add(member)
    database.session.commit()

    member = (
        Membership.query.options(joinedload(Membership.user))
        .filter_by(user_id=user.id, org_id=current_org.id)
        .one()
    )

    return member, 201


@appbp.route("/orgs/<aud:orgname>/roles")
@require_auth("org", "manage")
@json
def list_roles(orgname):
    roles = Role.query.filter_by(org_id=current_org.id).all()
    return [model.export_data() for model in roles], 200


@appbp.route("/orgs/<aud:orgname>/roles/<int:id>")
@require_auth("org", "manage")
@json
def get_role(orgname, id):
    return Role.query.filter_by(org_id=current_org.id, id=id).first_or_404()


role_schema = generate_schema(Role, exclude=["org_id"])
role_schema["abilities"] = {
    "type": "list",
    "schema": {
        "type": "dict",
        "schema": {
            "action_id": {"type": "integer", "required": True},
            "resource_id": {"type": "integer", "required": True},
        },
    },
}


@appbp.route("/orgs/<aud:orgname>/roles", methods=["POST"])
@require_auth("org", "manage")
@json
def add_role(orgname):
    """Creates a new role for an organization.

    Here an example of a request payload.

    .. code-block:: json

        {
            "name": "My role",
            "description": "This is an important role",
            "abilities": [
                {"action_id": 1 "resource_id": 4},
                {"action_id": 3 "resource_id": 27},
            ]
        }

    The ``abilities`` property is optional.
    """

    payload = request.get_json()
    v = Validator(role_schema)

    if v.validate(payload) is False:
        abort(400, v.errors)

    payload["org_id"] = current_org.id

    model = Role()
    model.import_data(payload)
    database.session.add(model)

    if "abilities" in payload:
        database.session.flush()

        for data in payload["abilities"]:

            data["role_id"] = model.id
            role_ability = RoleAbility()
            role_ability.import_data(data)
            database.session.add(role_ability)

    database.session.commit()

    model.abilities

    return model, 201


@appbp.route("/orgs/<aud:orgname>/roles/<int:id>", methods=["PATCH"])
@require_auth("org", "manage")
@json
def update_role(orgname, id):
    """ Updates an organization role. This endpoint also adds and removes
    abilities, by passing a list with the abilities that you want to add and
    omitting those that you want to remove. For instance, having the next role;

    .. code-block:: json

        {
            "name": "My role",
            "description": "This is a crazy role",
            "abilities": [
                {"action_id": 1, "resource_id": 4},
                {"action_id": 3, "resource_id": 27},
            ]
        }

    If I send a request as belog:

    .. code-block:: json

        {
            "name": "My new name",
            "abilities": [
                {"action_id": 1, "resource_id": 4},
                {"action_id": 4, "resource_id": 100},
            ]
        }

    This will result in ``{"action_id": 4 "resource_id": 100}`` being
    added and ``{"action_id": 3 "resource_id": 27}`` being removed.
    """

    payload = request.get_json()
    v = Validator(role_schema)

    if v.validate(payload, Role, update=True) is False:
        abort(400, v.errors)

    model = Role.query.filter_by(org_id=current_org.id, id=id).first_or_404()
    model.import_data(payload)

    # Here we checks if an ability is being removed or added
    if "abilities" in payload and isinstance(payload["abilities"], list):

        def was_removed(b):
            for index, value in enumerate(payload["abilities"]):
                # if the ability is still present we remove it from the request
                # value and return a False.
                if (
                    b.resource_id == value["resource_id"]
                    and b.action_id == value["action_id"]
                ):
                    payload["abilities"].pop(index)
                    return False
            return True

        # first we checks if an ability was removed
        for a in model.abilities:
            if was_removed(a):
                database.session.delete(a)

        # secondly we save the remaining abilities in the request value
        for c in payload["abilities"]:
            c["role_id"] = model.id
            role_ability = RoleAbility()
            role_ability.import_data(c)
            database.session.add(role_ability)

    database.session.commit()
    return model


@appbp.route("/orgs/<aud:orgname>/roles/<int:id>", methods=["DELETE"])
@require_auth("org", "manage")
@json
def delete_role(orgname, id):
    model = Role.query.filter_by(org_id=current_org.id, id=id).first_or_404()

    database.session.delete(model)
    database.session.commit()
    return {}


#  Member roles
#  ~~~~~~~~~~~~


def _verify_membership(username):
    # Check membership
    subquery = (
        database.session.query(User.id)
        .filter_by(canonical_username=username.lower())
        .subquery()
    )

    return Membership.query.filter_by(
        user_id=subquery.c.id, org_id=current_org.id
    ).first_or_404()


@appbp.route("/orgs/<aud:orgname>/members/<username>/roles")
@require_auth("org", "manage")
@json
def list_member_roles(orgname, username):
    """Returns all the roles assigned to a member."""

    member = _verify_membership(username)

    # Get roles
    member_role = (
        MemberRole.query.options(joinedload(MemberRole.role).joinedload(Role.abilities))
        .filter_by(user_id=member.user_id, org_id=member.org_id)
        .all()
    )

    return [item.role.export_data() for item in member_role], 200


@appbp.route("/orgs/<aud:orgname>/members/<username>/roles/<int:id>")
@require_auth("org", "manage")
@json
def get_member_role(orgname, username, id):
    """This endpoint serves to check if a member has a role assigned to him and
    get information about it as well.
    """

    member = _verify_membership(username)

    ident = (member.org_id, member.user_id, id)

    member_role = MemberRole.query.options(
        joinedload(MemberRole.role).joinedload(Role.abilities)
    ).get_or_404(ident)

    return member_role.role


@appbp.route("/orgs/<aud:orgname>/members/<username>/roles", methods=["POST"])
@require_auth("org", "manage")
@json
def add_member_role(orgname, username):
    """Assigns one or more role to a member.

    The request payload must have a **roles** property containing a list of
    role IDs that you want to assign. Don't send a role ID that already was
    assigned. Example::

        {"roles": [221, 33, 4000]}

    """

    payload = request.get_json()

    v = Validator(
        {"roles": {"type": "list", "minlength": 1, "schema": {"type": "integer"}}}
    )

    if v.validate(payload) is False:
        abort(400)

    member = _verify_membership(username)

    data = {"org_id": member.org_id, "user_id": member.user_id}

    for role_id in payload["roles"]:
        data["role_id"] = role_id
        member_role = MemberRole()
        member_role.import_data(data)
        database.session.add(member_role)

    database.session.commit()

    return {}, 201


@appbp.route(
    "/orgs/<aud:orgname>/members/<username>/roles/<int:id>", methods=["DELETE"]
)
@require_auth("org", "manage")
@json
def remove_member_role(orgname, username, id):
    """Remove a single role from a member."""

    member = _verify_membership(username)

    ident = (member.org_id, member.user_id, id)

    role = MemberRole.query.get_or_404(ident)

    database.session.delete(role)
    database.session.commit()

    return {}, 200
